#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2016-05-31
#
# monitor Check Point appliance soft RAID
#
# 2018-03-08: changed snmp scan function
# 2018-03-13: code cleanup
# 2018-03-15: more code cleanup
# 2020-06-08: changed snmp-scan function
# 2020-11-10: removed perfdata
#             added wato option to configure 'ignore disks'. For example on the Smart-1 5150 (Array has 12 Disks,
#             but only 6 Disks are in the default hardware
# 2021-04-29: rename from checkpoint_soft_raid in to checkpoint_raid
# 2021-08-10: rewritten for CMK 2.0
# 2023-04-23: refactoring
#
# sample snmpwalk (one volume / two disks)
#
#  .1.3.6.1.4.1.2620.1.6.7.7.1.1.1.1.0 = INTEGER: 1
#  .1.3.6.1.4.1.2620.1.6.7.7.1.1.2.1.0 = INTEGER: 0
#  .1.3.6.1.4.1.2620.1.6.7.7.1.1.3.1.0 = INTEGER: 2
#  .1.3.6.1.4.1.2620.1.6.7.7.1.1.4.1.0 = INTEGER: 2
#  .1.3.6.1.4.1.2620.1.6.7.7.1.1.5.1.0 = INTEGER: 1952448512
#  .1.3.6.1.4.1.2620.1.6.7.7.1.1.6.1.0 = INTEGER: 0
#  .1.3.6.1.4.1.2620.1.6.7.7.1.1.7.1.0 = INTEGER: 1
#  .1.3.6.1.4.1.2620.1.6.7.7.1.1.8.1.0 = Gauge32: 931
#  .1.3.6.1.4.1.2620.1.6.7.7.2.1.1.1.0 = INTEGER: 1
#  .1.3.6.1.4.1.2620.1.6.7.7.2.1.1.2.0 = INTEGER: 2
#  .1.3.6.1.4.1.2620.1.6.7.7.2.1.2.1.0 = INTEGER: 0
#  .1.3.6.1.4.1.2620.1.6.7.7.2.1.2.2.0 = INTEGER: 0
#  .1.3.6.1.4.1.2620.1.6.7.7.2.1.3.1.0 = INTEGER: 0
#  .1.3.6.1.4.1.2620.1.6.7.7.2.1.3.2.0 = INTEGER: 1
#  .1.3.6.1.4.1.2620.1.6.7.7.2.1.4.1.0 = INTEGER: 0
#  .1.3.6.1.4.1.2620.1.6.7.7.2.1.4.2.0 = INTEGER: 1
#  .1.3.6.1.4.1.2620.1.6.7.7.2.1.5.1.0 = STRING: "ATA     "
#  .1.3.6.1.4.1.2620.1.6.7.7.2.1.5.2.0 = STRING: "ATA     "
#  .1.3.6.1.4.1.2620.1.6.7.7.2.1.6.1.0 = STRING: "ST1000NM0033-9ZM"
#  .1.3.6.1.4.1.2620.1.6.7.7.2.1.6.2.0 = STRING: "ST1000NM0033-9ZM"
#  .1.3.6.1.4.1.2620.1.6.7.7.2.1.7.1.0 = STRING: "SN04"
#  .1.3.6.1.4.1.2620.1.6.7.7.2.1.7.2.0 = STRING: "SN04"
#  .1.3.6.1.4.1.2620.1.6.7.7.2.1.8.1.0 = INTEGER: 1953525168
#  .1.3.6.1.4.1.2620.1.6.7.7.2.1.8.2.0 = INTEGER: 1953525168
#  .1.3.6.1.4.1.2620.1.6.7.7.2.1.9.1.0 = INTEGER: 0
#  .1.3.6.1.4.1.2620.1.6.7.7.2.1.9.2.0 = INTEGER: 0
#  .1.3.6.1.4.1.2620.1.6.7.7.2.1.10.1.0 = INTEGER: 0
#  .1.3.6.1.4.1.2620.1.6.7.7.2.1.10.2.0 = INTEGER: 0
#  .1.3.6.1.4.1.2620.1.6.7.7.2.1.11.1.0 = INTEGER: 100
#  .1.3.6.1.4.1.2620.1.6.7.7.2.1.11.2.0 = INTEGER: 100
#  .1.3.6.1.4.1.2620.1.6.7.7.2.1.12.1.0 = Gauge32: 931
#  .1.3.6.1.4.1.2620.1.6.7.7.2.1.12.2.0 = Gauge32: 931
#
#
# sample info
#
# [[['1', '0', '2', '2', '1952448512', '0', '1', '931']],
#  [['1', '0', '0', '0', 'ATA     ', 'ST1000NM0033-9ZM', 'SN04', '1953525168', '0', '0', '100', '931'],
#   ['2', '0', '1', '1', 'ATA     ', 'ST1000NM0033-9ZM', 'SN04', '1953525168', '0', '0', '100', '931']]]
#
#

from dataclasses import dataclass
from typing import List, Dict
from cmk.base.plugins.agent_based.agent_based_api.v1.type_defs import (
    DiscoveryResult,
    CheckResult,
    StringTable,
)
from cmk.base.plugins.agent_based.agent_based_api.v1 import (
    register,
    Service,
    equals,
    Result,
    State,
    SNMPTree,
    startswith,
    all_of,
    any_of,
)


@dataclass
class RaidDisk:
    index: int
    volume_id: str
    disk_id: int
    vendor: str
    product_id: str
    revision: str
    max_lba: int
    state: int
    flags: str
    sync_state: int
    size: int


@dataclass
class RaidVolume:
    index: int
    type: str
    num_of_disks_on_raid: int
    max_lba: int
    state: int
    flags: str
    size: int
    disks: List[RaidDisk]
   

_raid_volume_type = {
    0: 'RAID-0',
    1: 'RAID-1E',
    2: 'RAID-1',
    3: 'RAID_10',
    4: 'RAID-4',
    5: 'RAID-5',
    6: 'RAID-6',
    7: 'RAID-60',
    8: 'RAID-50',
}


_raid_volume_state = {
    0: 'OPTIMAL',
    1: 'DEGRADED',
    2: 'FAILED',
}

_raid_volume_flags = {
    0: 'NONE',
    1: 'ENABLED',
    2: 'QUIESCED',
    3: 'RESYNC IN PROGRESS',
    4: 'VOLUME INACTIVE',
    5: 'NOT CONFIGURED',
    6: 'USING INTERIM RECOVERY MODE',
    7: 'READY FOR RECOVERY OPERATION',
    8: 'WRONG PHYSICAL DRIVE WAS REPLACED',
    9: 'A PHYSICAL DRIVE IS NOT PROPERLY CONNECTED',
    10: 'HARDWARE IS OVER HEATING',
    11: 'HARDWARE WAS OVERHEATED',
    12: 'CURRENTLY EXPENDING',
    13: 'NOT YET AVAILABLE',
    14: 'QUEUED FOR EXPENSION',
    15: 'MIGRATING',
    16: 'IMPACTED',
    17: 'OFFLINE',
    18: 'CLEARING',
}

_raid_disk_state = {
    0: 'ONLINE',
    1: 'MISSING',
    2: 'NOT COMPATIBLE',
    3: 'DISC FAILED',
    4: 'INITIALIZING',
    5: 'OFFLINE REQUESTED',
    6: 'FAILED REQUESTED',
    7: 'UNCONFIGURED GOOD SPUN UP',
    8: 'UNCONFIGURED GOOD SPUN DOWN',
    9: 'UNCONFIGURED BAD',
    10: 'HOTSPARE',
    11: 'DRIVE OFFLINE',
    12: 'REBUILD',
    13: 'FAILED',
    14: 'COPYBACK',
    255: 'OTHER OFFLINE',
}

_raid_disk_flags = {
    0: 'NONE',
    1: 'OUT OF SYNC',
    2: 'QUIESCED',
    3: 'VERIFYING',
    4: 'READY',
}


def _get_flags(flags, flags_description):
    description = ''
    flags = int(flags)
    if flags == 0:
        return flags_description.get(0)
    else:
        for x in range(0, len(flags_description)):
            if flags & (1 << x) != 0:
                if flags_description.get(x + 1) is not None:
                    description += f'{flags_description.get(x + 1)},'
        return description


def parse_checkpoint_raid(string_table: List[StringTable]) -> Dict[str, RaidVolume]:

    raw_volumes, raw_disks = string_table
    volumes = {}
    disks = []

    for disk in raw_disks:
        index, volume_id, disk_id, number, vendor, product_id, revision, max_lba, state, flags, sync_state, size = disk

        if index.isdigit():
            disks.append(RaidDisk(
                index=int(index),
                volume_id=volume_id.strip(),
                disk_id=int(number),
                vendor=vendor.strip(),
                product_id=product_id.strip(),
                revision=revision.strip(),
                max_lba=int(max_lba),
                state=int(state),
                flags=_get_flags(flags, _raid_disk_flags),
                sync_state=int(sync_state),
                size=int(size)
            ))

    for volume in raw_volumes:
        index, volume_id, _type, num_of_disks_on_raid, max_lba, state, flags, size = volume

        if index.isdigit():
            volumes[volume_id] = RaidVolume(
                index=int(index),
                type=_raid_volume_type.get(int(_type), f'unknown ({_type}'),
                num_of_disks_on_raid=int(num_of_disks_on_raid),
                max_lba=int(max_lba),
                state=int(state),
                flags=_get_flags(flags, _raid_volume_flags),
                size=int(size),
                disks=[]
            )
            for disk in disks:
                if disk.volume_id == volume_id:
                    volumes[volume_id].disks.append(disk)

    return volumes


def discovery_checkpoint_raid(section: Dict[str, RaidVolume]) -> DiscoveryResult:
    for key in section.keys():
        yield Service(item=key)


def check_checkpoint_raid(item, params, section: Dict[str, RaidVolume]) -> CheckResult:
    ignore_disks = params['ignore_disks'].split(',')

    try:
        volume = section[item]
    except KeyError:
        yield Result(state=State.CRIT, summary='Volume not found in SNMP Data')
        return

    # volume info

    details = ''
    details += f'Volume Type: {volume.type}, '
    details += f'# of Disks: {volume.num_of_disks_on_raid}, '
    details += f'Size: {volume.size}GB, '
    details += f'max LBA: {volume.max_lba}, '
    details += f'Flags: {volume.flags}'

    summary = ''
    summary += f'{volume.type}, '
    summary += f'{volume.num_of_disks_on_raid} disks, '
    summary += f'{volume.size}GB, '
    summary += f'State: {_raid_volume_state.get(volume.state)}'

    if volume.state != 0:
        yield Result(state=State.CRIT, summary=summary, details=details)
    else:
        yield Result(state=State.OK, summary=summary, details=details)

    # disk info
    disks = volume.disks
    disks_in_array = 0
    disks_online = 0
    disks_ignored = 0
    for disk in disks:
        if str(disk.disk_id) not in ignore_disks:
            disks_in_array += 1
            syncstate = disk.sync_state
            details = ''
            details += f'Disk ID: {disk.disk_id}, '
            details += f'State: {_raid_disk_state.get(disk.state)}, '
            details += f'Sync: {syncstate}%, '
            details += f'Size: {disk.size}GB, '
            if disk.max_lba > 0:
                details += f'max LBA: {disk.max_lba}, '
            if disk.vendor != 'UNKNOWN':
                details += f'Vendor: {disk.vendor}, '
            details += f'Product ID: {disk.product_id}, '
            if disk.revision != '':
                details += f'Revision: {disk.revision}, '
            details += f'Flags: {disk.flags}'
            if disk.state != 0:
                yield Result(state=State.CRIT,
                             notice=f'Disk ID: {disk.disk_id}, state: {_raid_disk_state.get(disk.state)}',
                             details=details)
            else:
                yield Result(state=State.OK, notice=details,
                             )
                disks_online += 1
        else:
            disks_ignored += 1

    if disks_in_array == disks_online:
        yield Result(state=State.OK, summary=f'{disks_online}/{disks_in_array} Disks online')
    else:
        yield Result(state=State.CRIT, summary=f'{disks_online}/{disks_in_array} Disks online')

    if disks_ignored > 0:
        yield Result(state=State.OK, summary=f'{disks_ignored} Disks ignored')


register.snmp_section(
    name='checkpoint_raid',
    parse_function=parse_checkpoint_raid,
    fetch=[
        SNMPTree(
            base='.1.3.6.1.4.1.2620.1.6.7.7.1.1',  # CHECKPOINT-MIB::raidVolumeEntry
            oids=[
                '1',  # raidVolumeIndex
                '2',  # raidVolumeID
                '3',  # raidVolumeType
                '4',  # numOfDisksOnRaid
                '5',  # raidVolumeMaxLBA
                '6',  # raidVolumeState
                '7',  # raidVolumeFlags
                '8',  # raidVolumeSize
            ]
        ),
        SNMPTree(
            base='.1.3.6.1.4.1.2620.1.6.7.7.2.1',  # CHECKPOINT-MIB::fwLSConnEntry
            oids=[
                '1',  # raidDiskIndex
                '2',  # raidDiskVolumeID
                '3',  # raidDiskID
                '4',  # raidDiskNumber
                '5',  # raidDiskVendor
                '6',  # raidDiskProductID
                '7',  # raidDiskRevision
                '8',  # raidDiskMaxLBA
                '9',  # raidDiskState
                '10',  # raidDiskFlags
                '11',  # raidDiskSyncState
                '12',  # raidDiskSize
            ]
        )

    ],
    detect=any_of(
        startswith('.1.3.6.1.2.1.1.2.0', '.1.3.6.1.4.1.2620'),
        all_of(
            equals('.1.3.6.1.2.1.1.2.0', '.1.3.6.1.4.1.8072.3.2.10'),
            equals('.1.3.6.1.4.1.2620.1.6.1.0', 'SVN Foundation'),
        )
    )
)

register.check_plugin(
    name='checkpoint_raid',
    service_name='RAID ID %s',
    discovery_function=discovery_checkpoint_raid,
    check_function=check_checkpoint_raid,
    check_default_parameters={
        'ignore_disks': ''
    },
    check_ruleset_name='checkpoint_raid',
)
